from dataclasses import dataclass
from pathlib import Path

from pvi._schema_utils import as_discriminated_union
from pvi.device import Device


@as_discriminated_union
@dataclass
class Formatter:
    def format(self, device: Device, prefix: str, path: Path):
        raise NotImplementedError(self)



from dataclasses import dataclass
from typing import Dict, List, Optional

from pvi._produce.asyn import AsynParameter


@dataclass
class Record:
    name: str  # The name of the record e.g. $(P)$(M)Status
    type: str  # The record type string e.g. ao, stringin
    fields: Dict[str, str]  # The record fields
    infos: Dict[str, str]  # Any infos to be added to the record


class Parameter:
    invalid = ["DESC", "DTYP", "INP", "OUT", "PINI", "VAL"]

    def _remove_invalid(self, fields: Dict[str, str]) -> Dict[str, str]:
        valid_fields = {
            key: value for (key, value) in fields.items() if key not in self.invalid
        }
        return valid_fields

    def generate_component(self) -> AsynParameter:
        raise NotImplementedError(self)


class ReadParameterMixin:
    def _get_read_record_suffix(self) -> Optional[str]:
        raise NotImplementedError(self)


class WriteParameterMixin:
    def _get_initial(self, write_record: Record) -> str:
        return write_record.fields.get("VAL", "")

    def _get_demand_auto_updates(self, write_record: Record) -> bool:
        return write_record.infos.get("asyn:READBACK") == "1"

    def _get_autosave_fields(self, write_record: Record) -> List[str]:
        autosave_fields = write_record.infos.get("autosaveFields")
        if autosave_fields:
            autosave_list = autosave_fields.split(" ")
        else:
            autosave_list = []
        return autosave_list



import csv
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, Iterator, List, Optional, Union

from apischema.schemas import schema
from apischema.types import Number
from epicsdbbuilder import WriteRecords, records
from epicsdbbuilder.recordbase import Record
from typing_extensions import Annotated

from pvi._schema_utils import as_discriminated_union, desc
from pvi._yaml_utils import deserialize_yaml
from pvi.device import (
    LED,
    CheckBox,
    ComboBox,
    Component,
    Device,
    Group,
    Named,
    ReadWidget,
    SignalR,
    SignalRW,
    SignalW,
    TextRead,
    TextWrite,
    Tree,
    WriteWidget,
    on_each_node,
    walk,
)

from .base import Access, DisplayForm, Producer
from .records import (
    PVI_NELM,
    AnalogueRecordPair,
    BinaryRecordPair,
    BusyRecordPair,
    LongRecordPair,
    MultiBitBinaryRecordPair,
    PVIRecord,
    RecordPair,
    StringRecordPair,
    WaveformRecordPair,
)
from .utils import get_param_set, join_lines, truncate_description


@dataclass
class TypeStrings:
    """The type strings for record dtypes and parameter names"""

    asyn_read: Annotated[str, desc("e.g. asynInt32, asynOctetRead")]
    asyn_write: Annotated[str, desc("e.g. asynInt32, asynOctetWrite")]
    asyn_param: Annotated[str, desc("e.g. asynParamInt32, asynParamOctet")]


AReadWidget = Annotated[Optional[ReadWidget], desc("Widget to use for read record")]
AWriteWidget = Annotated[Optional[WriteWidget], desc("Widget to use for write record")]


@as_discriminated_union
@dataclass
class AsynParameter(Named):
    """Base class for all Asyn Parameters to inherit from"""

    description: Annotated[str, desc("Description of what this Parameter is for")]
    index_name: Annotated[
        Optional[str],
        desc("Override name of index variable in source code (defaults to name)"),
    ] = None
    drv_info: Annotated[
        Optional[str],
        desc(
            "Info string for drvUserCreate for dynamically created parameters",
            # Limit to 40 characters with no whitespace
            pattern=r"^\S{1,40}$",
        ),
    ] = None
    access: Annotated[Access, desc(Access.__doc__)] = Access.RW
    demand_auto_updates: Annotated[
        bool, desc("Should demand update when readback changes?")
    ] = False
    read_record_suffix: Annotated[
        Optional[str], desc("The read record suffix, if not given then use $(name)_RBV")
    ] = None
    write_record_suffix: Annotated[
        Optional[str], desc("The write record suffix, if not given then use $(name)")
    ] = None
    display_form: Annotated[
        Optional[DisplayForm], desc("Display form for numeric/array fields")
    ] = None
    type_strings: ClassVar[TypeStrings]
    initial: Any = None
    read_widget: Optional[AReadWidget] = None
    write_widget: Optional[AWriteWidget] = None
    record_fields: RecordPair = RecordPair()

    def get_index_name(self):
        return self.index_name or self.name

    def get_drv_info(self):
        return self.drv_info or self.name


def initial_value(pattern: str = None, min: Number = None, max: Number = None):
    if pattern:
        pattern = r"[\'\"]?(?:\$\(\w+=)?" + pattern + r"\)?[\'\"]?"
    return schema(
        description="The initial value of the parameter",
        min=min,
        max=max,
        pattern=pattern,
    )


# The child classes of AsynParameter redefine record_fields with a dynamic type to
# generate a schema with fields specific to the parameter type. However, the dymanic
# types make mypy unhappy when it tries to do static type checking, so we use
# `type: ignore`.
# https://mypy.readthedocs.io/en/stable/common_issues.html#variables-vs-type-aliases


@dataclass
class AsynBinary(AsynParameter):
    """Asyn Binary Parameter and records"""

    type_strings = TypeStrings(
        asyn_read="asynInt32",
        asyn_write="asynInt32",
        asyn_param="asynParamInt32",
    )
    initial: Annotated[
        Union[None, int, str], initial_value(r"[0-1]", min=0, max=1)
    ] = None
    read_widget: AReadWidget = LED()
    write_widget: AWriteWidget = CheckBox()
    record_fields: Annotated[  # type: ignore
        BinaryRecordPair, desc("Binary record fields")
    ] = BinaryRecordPair()


@dataclass
class AsynBusy(AsynBinary):
    """Asyn Busy Parameter and records"""

    record_fields: Annotated[  # type: ignore
        BusyRecordPair, desc("Busy record fields")
    ] = BusyRecordPair()


@dataclass
class AsynFloat64(AsynParameter):
    """Asyn Float64 Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynFloat64",
        asyn_write="asynFloat64",
        asyn_param="asynParamFloat64",
    )
    initial: Annotated[
        Union[None, float, str], initial_value(r"-?[0-9]+(\.[0-9]+)?")
    ] = None
    read_widget: AReadWidget = TextRead()
    write_widget: AWriteWidget = TextWrite()
    record_fields: Annotated[  # type: ignore
        AnalogueRecordPair, desc("Analogue record fields")
    ] = AnalogueRecordPair()


@dataclass
class AsynInt32(AsynParameter):
    """Asyn Int32 Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynInt32",
        asyn_write="asynInt32",
        asyn_param="asynParamInt32",
    )
    initial: Annotated[Union[None, int, str], initial_value(r"-?\d+")] = None
    read_widget: AReadWidget = TextRead()
    write_widget: AWriteWidget = TextWrite()
    record_fields: Annotated[  # type: ignore
        AnalogueRecordPair, desc("Analogue record fields")
    ] = AnalogueRecordPair()


@dataclass
class AsynLong(AsynInt32):
    """Asyn Long Parameter and records"""

    record_fields: Annotated[  # type: ignore
        LongRecordPair, desc("Long record fields")
    ] = LongRecordPair()


@dataclass
class AsynMultiBitBinary(AsynParameter):
    """Asyn MultiBitBinary Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynInt32",
        asyn_write="asynInt32",
        asyn_param="asynParamInt32",
    )
    initial: Annotated[
        # Regex: [0-9] OR 1[0-5] -> [0-15]  (Single character matching)
        Union[None, int, str],
        initial_value(r"([0-9]|1[0-5])", min=0, max=15),
    ] = None
    read_widget: AReadWidget = TextRead()
    write_widget: AWriteWidget = ComboBox()
    record_fields: Annotated[  # type: ignore
        MultiBitBinaryRecordPair, desc("Multi-bit binary record fields")
    ] = MultiBitBinaryRecordPair()


@dataclass
class AsynString(AsynParameter):
    """Asyn String Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynOctetRead",
        asyn_write="asynOctetWrite",
        asyn_param="asynParamOctet",
    )
    initial: Annotated[Union[None, str], initial_value()] = None
    read_widget: AReadWidget = TextRead()
    write_widget: AWriteWidget = TextWrite()
    record_fields: Annotated[  # type: ignore
        StringRecordPair, desc("String record fields")
    ] = StringRecordPair()


@dataclass
class AsynWaveform(AsynParameter):
    """Asyn Waveform Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynOctetRead",
        asyn_write="asynOctetWrite",
        asyn_param="asynParamOctet",
    )
    read_widget: AReadWidget = TextRead()
    write_widget: AWriteWidget = TextWrite()
    record_fields: Annotated[  # type: ignore
        WaveformRecordPair, desc("Waveform record fields")
    ] = WaveformRecordPair()


@dataclass
class AsynInt32Waveform(AsynWaveform):
    """Asyn Waveform Parameter and records with int32 array elements"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynInt32ArrayIn",
        asyn_write="asynInt32ArrayOut",
        asyn_param="asynParamInt32",
    )
    # We have to redefine this to retain the `type: ignore` for mypy
    record_fields: Annotated[  # type: ignore
        WaveformRecordPair, desc("Waveform record fields")
    ] = WaveformRecordPair()


@dataclass
class AsynFloat64Waveform(AsynWaveform):
    """Asyn Waveform Parameter and records with int32 array elements"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynFloat64ArrayIn",
        asyn_write="asynFloat64ArrayOut",
        asyn_param="asynParamFloat64",
    )
    # We have to redefine this to retain the `type: ignore` for mypy
    record_fields: Annotated[  # type: ignore
        WaveformRecordPair, desc("Waveform record fields")
    ] = WaveformRecordPair()


def find_components(yaml_name: str, yaml_paths: List[Path]) -> Tree[AsynParameter]:
    if yaml_name == "asynPortDriver":
        return []  # asynPortDriver is the most base class and has no parameters

    # Look in this module first
    producer_name = f"{yaml_name}.pvi.producer.yaml"
    producer_yaml = find_pvi_yaml(producer_name, yaml_paths)

    if producer_yaml is None:
        raise IOError(f"Cannot find {producer_name} in {yaml_paths}")

    producer = deserialize_yaml(AsynProducer, producer_yaml)

    return list(producer.parameters) + list(
        find_components(producer.parent, yaml_paths)
    )


def find_pvi_yaml(yaml_name: str, yaml_paths: List[Path]) -> Union[Path, None]:
    """Find a yaml file in given directory"""
    for yaml_path in yaml_paths:
        if yaml_path.is_dir():
            if yaml_name in [f.name for f in yaml_path.iterdir()]:
                return yaml_path / yaml_name
    return None


@dataclass
class AsynProducer(Producer):
    label: Annotated[str, desc("Screen title")]
    asyn_port: Annotated[str, desc("The asyn port name")]
    address: Annotated[str, desc("The asyn address")]
    timeout: Annotated[str, desc("The timeout for the asyn port")]
    parent: Annotated[
        str,
        desc(
            "The parent producer (basename of yaml file), "
            "asynPortDriver is the top of the tree"
        ),
    ]
    parameters: Annotated[
        Tree[AsynParameter], desc("The parameters to make into an IOC")
    ]

    def deserialize_parents(self, yaml_paths: List[Path]):
        """Deserialize yaml of parents and extract parameters"""
        if self.parent == "asynPortDriver":
            pass

        parent_parameters = find_components(self.parent, yaml_paths)
        for node in parent_parameters:
            if isinstance(node, Group):
                for param_group in self.parameters:
                    if not isinstance(param_group, Group):
                        continue
                    elif param_group.name == node.name:
                        param_group.children = list(node.children) + list(
                            param_group.children
                        )
                        break  # Groups merged - skip to next parent group

                else:  # No break - Did not find the Group
                    # Inherit as a new Group
                    self.parameters = list(self.parameters) + [node]
                    continue  # Skip to next parent group

            else:
                # Node is an individual AsynParameter - just append it
                self.parameters = list(self.parameters) + [node]

    def _read_record_suffix(self, parameter: AsynParameter) -> str:
        if parameter.read_record_suffix:
            return parameter.read_record_suffix
        else:
            return parameter.name + "_RBV"

    def _write_record_suffix(self, parameter: AsynParameter) -> str:
        if parameter.write_record_suffix:
            return parameter.write_record_suffix
        else:
            return parameter.name

    def produce_device(self) -> Device:
        """Make signals from components"""
        components = on_each_node(self.parameters, self._produce_component)
        return Device(self.label, components)

    def _produce_component(self, parameter: AsynParameter) -> Iterator[Component]:
        # TODO: what about SignalX?
        read_pv = self._read_record_suffix(parameter)
        write_pv = self._write_record_suffix(parameter)
        if parameter.access == Access.R:
            yield SignalR(parameter.name, read_pv, parameter.read_widget)
        elif parameter.access == Access.W:
            yield SignalW(parameter.name, write_pv, parameter.write_widget)
        elif parameter.access == Access.RW:
            need_both = not parameter.demand_auto_updates
            yield SignalRW(
                parameter.name,
                write_pv,
                parameter.write_widget,
                read_pv=read_pv if need_both else "",
                read_widget=parameter.read_widget if need_both else None,
            )

    def produce_csv(self, path: Path):
        with open(path, "w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=",", quotechar='"')
            writer.writerow(
                [
                    "Parameter Index Variable",
                    "Asyn Interface",
                    "Access",
                    "drvInfo String",
                    "Record Names",
                    "Record Types",
                    "Description",
                ]
            )
            for node in walk(self.parameters):
                if isinstance(node, Group):
                    writer.writerow([f"*{node.name}*"] + [""] * 6)
                else:
                    names = []
                    types = []
                    interfaces = []
                    if node.access.needs_write_record():
                        names.append(self.prefix + self._write_record_suffix(node))
                        types.append(node.record_fields.out_record_type.__name__)
                        interfaces.append(node.type_strings.asyn_write)
                    if node.access.needs_read_record():
                        names.append(self.prefix + self._read_record_suffix(node))
                        types.append(node.record_fields.in_record_type.__name__)
                        if node.type_strings.asyn_read not in interfaces:
                            interfaces.append(node.type_strings.asyn_read)

                    writer.writerow(
                        [
                            node.index_name,
                            ",\n".join(interfaces),
                            node.access.name,
                            node.drv_info,
                            ",\n".join(names),
                            ",\n".join(types),
                            node.description,
                        ]
                    )

    def produce_records(self, path: Path):
        """Make epicsdbbuilder records"""
        comment = None
        for node in walk(self.parameters):
            if isinstance(node, Group):
                comment = f"Group: {node.name}"
            else:
                for record in self._produce_record(node):
                    if node.display_form:
                        record.add_info("Q:form", node.display_form)
                    if comment:
                        record.add_comment(comment)
                    comment = None

        if self.parent == "asynPortDriver":
            PVIRecord(self.prefix + "PVI")

        header = """\
This file was automatically generated
*** Please do not edit this file: edit the source file instead. ***
"""
        WriteRecords(str(path), header=header, alphabetical=False)

    def _produce_record(self, parameter: AsynParameter) -> Iterator[Record]:
        inp_fields, out_fields = parameter.record_fields.sort_records()
        drv_info = parameter.get_drv_info()
        io = f"@asyn({self.asyn_port},{self.address},{self.timeout}){drv_info}"
        if parameter.access.needs_read_record():
            name = self.prefix + self._read_record_suffix(parameter)
            rtype = parameter.record_fields.in_record_type.__name__
            inp_fields.update(
                DESC=truncate_description(parameter.description),
                DTYP=parameter.type_strings.asyn_read,
                INP=io,
            )
            yield getattr(records, rtype)(name, **inp_fields)
        if parameter.access.needs_write_record():
            name = self.prefix + self._write_record_suffix(parameter)
            rtype = parameter.record_fields.out_record_type.__name__
            out_fields.update(
                DESC=truncate_description(parameter.description),
                DTYP=parameter.type_strings.asyn_write,
            )
            if rtype == "waveform":
                out_fields["INP"] = io
            else:
                out_fields["OUT"] = io
            if parameter.initial is not None:
                out_fields.update(PINI="YES", VAL=parameter.initial)
            out_fields.pop("SCAN", None)
            record = getattr(records, rtype)(name, **out_fields)
            if parameter.demand_auto_updates:
                record.add_info("asyn:READBACK", 1)
            yield record

    def produce_other(self, path: Path, yaml_paths: List[Path]):
        """Make things like cpp, h files"""
        is_valid = path.suffix == ".h" and "." not in path.stem
        parent_param_set = get_param_set(self.parent)
        assert is_valid, f"Can only make header files not {path}"
        guard_define = path.stem[0].upper() + path.stem[1:] + "_H"
        defines = []
        adds = []
        indexes = []

        # Special case to add NDArrayData as first asynNDArrayDriver parameter
        # There is also a special case to remove it from the source files
        if path.name.startswith("asynNDArrayDriver"):
            defines.append('#define NDArrayDataString "ARRAY_DATA"')
            adds.append(
                "this->add(NDArrayDataString, asynParamGenericPointer, &NDArrayData);"
            )
            indexes.append("int NDArrayData;")
            indexes.append(f"#define FIRST_{path.stem.upper()}_PARAM NDArrayData")

        for node in walk(self.parameters):
            if isinstance(node, AsynParameter):
                param_type = node.type_strings.asyn_param
                index_name = node.get_index_name()

                defines.append(f'#define {index_name}String "{node.get_drv_info()}"')
                adds.append(
                    f"this->add({index_name}String, {param_type}, &{index_name});"
                )
                indexes.append(f"int {index_name};")

                if len(indexes) == 1:
                    indexes.append(
                        f"#define FIRST_{path.stem.upper()}_PARAM {index_name}"
                    )

        self.deserialize_parents(yaml_paths)
        param_tree = self.produce_device().generate_param_tree()
        assert len(param_tree) <= PVI_NELM, "Param tree limited to 100000 characters"

        path.write_text(
            f"""\
#ifndef {guard_define}
#define {guard_define}

#include "{parent_param_set}.h"

{join_lines(defines)}

const std::string {self.label}ParamTree = \\
{param_tree};

class {path.stem} : public virtual {parent_param_set} {{
public:
    {path.stem}() {{
        {join_lines(adds, indent=8)}

        this->paramTree = {self.label}ParamTree;
    }}

    {join_lines(indexes, indent=4)}
}};

#endif // {guard_define}
"""
        )



from dataclasses import dataclass
from pathlib import Path
from typing import List

from typing_extensions import Annotated

from pvi._schema_utils import desc
from pvi.device import Device

from .base import Formatter
from .utils import (
    ActionFactory,
    AdlTemplate,
    Bounds,
    GroupFactory,
    GroupType,
    LabelFactory,
    LayoutProperties,
    PVWidgetFactory,
    Screen,
    ScreenWidgets,
    WidgetFactory,
    with_title,
)


@dataclass
class APSFormatter(Formatter):
    spacing: Annotated[int, desc("Spacing between widgets")] = 5
    title_height: Annotated[int, desc("Height of screen title bar")] = 25
    max_height: Annotated[int, desc("Max height of the screen")] = 900
    label_width: Annotated[int, desc("Width of the widget description labels")] = 205
    widget_width: Annotated[int, desc("Width of the widgets")] = 100
    widget_height: Annotated[int, desc("Height of the widgets")] = 20

    def format(self, device: Device, prefix: str, path: Path):
        assert path.suffix == ".adl", "Can only write adl files"
        template = AdlTemplate((Path(__file__).parent / "aps.adl").read_text())
        layout = LayoutProperties(
            spacing=self.spacing,
            title_height=self.title_height,
            max_height=self.max_height,
            group_label_height=25,
            label_width=self.label_width,
            widget_width=self.widget_width,
            widget_height=self.widget_height,
            group_widget_indent=0,
            group_width_offset=0,
        )
        screen_widgets = ScreenWidgets(
            label_cls=LabelFactory.from_template(
                template, search='"Label"', textix="text"
            ),
            led_cls=PVWidgetFactory.from_template(
                template, search='"LED"', sized=Bounds.square, chan="pv"
            ),
            progress_bar_cls=PVWidgetFactory.from_template(
                template, search='"ProgressBar"', chan="pv"
            ),
            text_read_cls=PVWidgetFactory.from_template(
                template, search='"TextRead"', chan="pv"
            ),
            check_box_cls=PVWidgetFactory.from_template(
                template, search='"CheckBox"', chan="pv"
            ),
            combo_box_cls=PVWidgetFactory.from_template(
                template, search='"ComboBox"', chan="pv"
            ),
            text_write_cls=PVWidgetFactory.from_template(
                template, search='"TextWrite"', chan="pv"
            ),
            action_button_cls=ActionFactory.from_template(
                template, search='"SignalX"', label="label", chan="pv"
            ),
        )

        label_background_cls = group_box_cls = WidgetFactory.from_template(
            template, search="clr=2"
        )
        screen_title_cls = LabelFactory.from_template(
            template, search='"Title"', textix="text"
        )
        group_title_cls = LabelFactory.from_template(
            template, search='"Group"', textix="text"
        )
        group_box_cls = WidgetFactory.from_template(template, search='fill="outline"')

        def make_group_widgets(bounds: Bounds, title: str) -> List[WidgetFactory[str]]:
            title_bounds = Bounds(
                bounds.x + layout.spacing,
                bounds.y + layout.spacing,
                bounds.w - 2 * layout.spacing,
                layout.group_label_height - layout.spacing,
            )
            return [
                group_box_cls(bounds),
                label_background_cls(title_bounds),
                group_title_cls(title_bounds, title),
            ]

        def make_screen_widgets(bounds: Bounds, title: str) -> List[WidgetFactory[str]]:
            title_bounds = Bounds(0, 0, bounds.w, layout.title_height)
            return [
                label_background_cls(title_bounds),
                screen_title_cls(title_bounds, title),
            ]

        screen = Screen(
            screen_cls=GroupFactory.from_template(
                template,
                search=GroupType.SCREEN,
                sized=with_title(layout.spacing, layout.title_height),
                make_widgets=make_screen_widgets,
            ),
            group_cls=GroupFactory.from_template(
                template,
                search=GroupType.GROUP,
                sized=with_title(layout.spacing, layout.group_label_height),
                make_widgets=make_group_widgets,
            ),
            screen_widgets=screen_widgets,
            prefix=prefix,
            layout=layout,
        )
        title = f"{device.label} - {prefix}"
        texts = screen.screen(device.children, title).format()
        path.write_text("".join(texts))



import re
import sys
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Type, cast

from pvi._produce.asyn import Access, AsynParameter, AsynWaveform
from pvi._schema_utils import rec_subclasses

from ._parameters import Parameter, ReadParameterMixin, Record, WriteParameterMixin


class RecordError(Exception):
    pass


@dataclass
class AsynRecord(Record):
    def __post_init__(self):
        # We don't care about records without INP or OUT or with both (error)
        if all(k in self.fields.keys() for k in ("INP", "OUT")) or not any(
            k in self.fields.keys() for k in ("INP", "OUT")
        ):
            raise RecordError("Record has no input or output field or both")

        asyn_field = self.fields.get("INP", self.fields.get("OUT"))
        if "@asyn(" not in asyn_field:
            # Is it Asyn, because if it's not we don't care about it
            raise RecordError("Record has no @asyn field")

        # If there is no DESC field we must create one
        if "DESC" not in self.fields.keys():
            self.fields["DESC"] = self.name

    def get_parameter_name(self) -> Optional[str]:
        # e.g. from: field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
        # extract: FILE_PATH
        parameter_name_extractor = r"(?:@asyn\()(?:(?:\$\([^\)]*\)[,]*)*)(?:\))(\S+)"
        parameter_name = None
        for k, v in self.fields.items():
            if k == "INP" or k == "OUT":
                if "@asyn(" in v:
                    match = re.search(parameter_name_extractor, v)
                    if match:
                        parameter_name = match.group(1)
        return parameter_name

    def asyn_component_type(self) -> Type[AsynParameter]:
        # For waveform records the data type is defined by DTYP
        if self.type == "waveform":
            waveform_components = [AsynWaveform] + cast(
                List[Type[AsynWaveform]], rec_subclasses(AsynWaveform)
            )
            for waveform_cls in waveform_components:
                if self.fields["DTYP"] in (
                    waveform_cls.type_strings.asyn_read,
                    waveform_cls.type_strings.asyn_write,
                ):
                    return waveform_cls
            assert False, f"Waveform type for {self} not found in {waveform_components}"

        asyn_components = cast(List[Type[AsynParameter]], rec_subclasses(AsynParameter))
        if "INP" in self.fields.keys():
            type_fields = [
                (
                    cls.record_fields.in_record_type.__name__,
                    cls.type_strings.asyn_read,
                )
                for cls in asyn_components
            ]
        elif "OUT" in self.fields.keys():
            type_fields = [
                (
                    cls.record_fields.out_record_type.__name__,
                    cls.type_strings.asyn_write,
                )
                for cls in asyn_components
            ]
        try:
            idx = type_fields.index((self.type, self.fields["DTYP"]))
            asyn_class = asyn_components[idx]
            return asyn_class
        except ValueError as e:
            raise ValueError(
                f"{self.name} asyn type: ({self.type}, {self.fields['DTYP']}) "
                f"not found in {type_fields}"
            ) from e
        except KeyError as e:
            raise KeyError(f"{self.name} DTYP not found") from e


@dataclass
class SettingPair(Parameter, WriteParameterMixin, ReadParameterMixin):
    read_record: AsynRecord
    write_record: AsynRecord
    _dominant_read = True

    @property
    def dominant(self):
        return self.read_record if self._dominant_read else self.write_record

    @property
    def subordinate(self):
        return self.write_record if self._dominant_read else self.read_record

    def _get_read_record_suffix(self) -> Optional[str]:
        if self.read_record.name != self.write_record.name + "_RBV":
            return self.read_record.name
        else:
            return None

    def _get_field_clashes(self) -> List[str]:
        # Check to see if there any clashing values in pairs
        clashing_fields = []
        for (read_field_name, read_field_value) in self.read_record.fields.items():
            write_field_value = self.write_record.fields.get(
                read_field_name, read_field_value
            )
            if (
                write_field_value != read_field_value
                and read_field_name not in self.invalid
            ):
                clashing_fields.append(read_field_name)
        return clashing_fields

    def _handle_clashes(self, clashing_fields: List[str]):
        for field in clashing_fields:
            chosen_value = self.dominant.fields[field]
            discarded_value = self.subordinate.fields.get(field, chosen_value)
            print(
                f"Pair: {self.write_record.name}; "
                f"Field: {field}; "
                f"Values: {chosen_value}, "
                f"{discarded_value}; "
                f"Using {chosen_value} "
                f"for both",
                file=sys.stderr,
            )
            self.subordinate.fields[field] = self.dominant.fields.get(
                field, chosen_value
            )

    def get_naming_overrides(self) -> Tuple[str, List[str]]:
        record_name = self.subordinate.name
        override_fields = self._get_field_clashes()
        return record_name, override_fields

    def has_clashes(self) -> bool:
        empty = not self._get_field_clashes()
        return not empty

    def generate_component(self) -> AsynParameter:
        asyn_class = self.write_record.asyn_component_type()

        non_default_args: Dict[str, Any] = dict()
        non_default_args["demand_auto_updates"] = self._get_demand_auto_updates(
            self.write_record
        )
        autosave = self._get_autosave_fields(self.write_record)
        if autosave:
            pass
            # TODO: Consider handling autosave fields - see Action.generate_component
            # non_default_args["autosave"] = autosave

        drv_info = self.write_record.get_parameter_name()
        if drv_info != self.write_record.name:
            non_default_args["drv_info"] = drv_info

        initial = self._get_initial(self.write_record)
        if initial:
            non_default_args["initial"] = initial

        read_record_suffix = self._get_read_record_suffix()
        if read_record_suffix is not None:
            non_default_args["read_record_suffix"] = read_record_suffix

        field_clashes = self._get_field_clashes()
        self._handle_clashes(field_clashes)

        write_fields = self._remove_invalid(self.write_record.fields)
        read_fields = self._remove_invalid(self.read_record.fields)
        record_class = type(asyn_class.record_fields)
        kwargs = {**write_fields, **read_fields}
        record = record_class(**kwargs)  # type: ignore

        component = asyn_class(
            description=self.write_record.fields["DESC"],
            name=self.write_record.name,
            access=Access.RW,
            record_fields=record,
            **non_default_args,
        )
        return component


@dataclass
class Readback(Parameter, ReadParameterMixin):
    read_record: AsynRecord

    def _get_read_record_suffix(self) -> Optional[str]:
        if not self.read_record.name.endswith("_RBV"):
            return self.read_record.name
        else:
            return None

    def generate_component(self) -> AsynParameter:
        asyn_class = self.read_record.asyn_component_type()

        non_default_args: Dict[str, Any] = dict()
        read_record_suffix = self._get_read_record_suffix()
        if read_record_suffix:
            name = self.read_record.name
            non_default_args["read_record_suffix"] = read_record_suffix
        else:
            name = self.read_record.name[: -len("_RBV")]

        drv_info = self.read_record.get_parameter_name()
        if drv_info != self.read_record.name:
            non_default_args["drv_info"] = drv_info

        read_fields = self._remove_invalid(self.read_record.fields)
        record = type(asyn_class.record_fields)(**read_fields)  # type: ignore
        component = asyn_class(
            description=self.read_record.fields["DESC"],
            name=name,
            access=Access.R,
            record_fields=record,
            **non_default_args,
        )
        return component


@dataclass
class Action(Parameter, WriteParameterMixin):
    write_record: AsynRecord

    def generate_component(self) -> AsynParameter:
        asyn_class = self.write_record.asyn_component_type()

        non_default_args: Dict[str, Any] = dict()
        non_default_args["demand_auto_updates"] = self._get_demand_auto_updates(
            self.write_record
        )
        autosave_fields = self._get_autosave_fields(self.write_record)
        if autosave_fields:
            print(
                "Warning: Ignoring autosave fields. Consider how to handle this",
                file=sys.stderr,
            )
            # non_default_args["autosave"] = autosave_fields

        initial = self._get_initial(self.write_record)
        if initial:
            non_default_args["initial"] = initial

        drv_info = self.write_record.get_parameter_name()
        if drv_info != self.write_record.name:
            non_default_args["drv_info"] = drv_info

        write_fields = self._remove_invalid(self.write_record.fields)
        record = type(asyn_class.record_fields)(**write_fields)  # type: ignore
        component = asyn_class(
            description=self.write_record.fields["DESC"],
            name=self.write_record.name,
            access=Access.W,
            record_fields=record,
            **non_default_args,
        )
        return component



from __future__ import annotations

import re
from copy import deepcopy
from dataclasses import dataclass, field, replace
from enum import Enum
from typing import Callable, Dict, Iterator, List, Tuple, Type, TypeVar, Union

from lxml import etree
from typing_extensions import Annotated

from pvi._schema_utils import desc
from pvi.device import (
    LED,
    CheckBox,
    ComboBox,
    Component,
    Generic,
    Grid,
    Group,
    ProgressBar,
    ReadWidget,
    SignalR,
    SignalRef,
    SignalRW,
    SignalW,
    SignalX,
    TextRead,
    TextWrite,
    Tree,
    WriteWidget,
)

T = TypeVar("T")


@dataclass
class Bounds:
    x: int = 0
    y: int = 0
    w: int = 0
    h: int = 0

    def copy(self) -> Bounds:
        return Bounds(self.x, self.y, self.w, self.h)

    def split(self, width: int, spacing: int) -> Tuple[Bounds, Bounds]:
        """Split horizontally"""
        to_split = width + spacing
        assert to_split < self.w, f"Can't split off {to_split} from {self.w}"
        left = Bounds(self.x, self.y, width, self.h)
        right = Bounds(self.x + to_split, self.y, self.w - to_split, self.h)
        return left, right

    def square(self) -> Bounds:
        """Return the largest square that will fit in self"""
        size = min(self.w, self.h)
        return Bounds(
            x=self.x + int((self.w - size) / 2),
            y=self.y + int((self.h - size) / 2),
            w=size,
            h=size,
        )

    def added_to(self, bounds: Bounds) -> Bounds:
        return Bounds(
            x=self.x + bounds.x,
            y=self.y + bounds.y,
            w=self.w + bounds.w,
            h=self.h + bounds.h,
        )


class WidgetTemplate(Generic[T]):
    screen: T

    def search(self, search) -> T:
        """Search for a widget"""
        raise NotImplementedError(self)

    def set(self, t: T, bounds: Bounds = None, **properties) -> T:
        """Return a copy of the internal representation with the bounds and
        properties set"""
        raise NotImplementedError(self)

    def create_group(
        self,
        group_object: List[T],
        children: List[WidgetFactory[T]],
        padding: Bounds = Bounds(),
    ) -> List[T]:
        """Return a group widget with its children attached and appropritately padded"""
        raise NotImplementedError(self)


WF = TypeVar("WF", bound="WidgetFactory")


@dataclass
class WidgetFactory(Generic[T]):
    bounds: Bounds

    def format(self) -> List[T]:
        """Will be filled in by from_template"""
        raise NotImplementedError(self)

    @classmethod
    def from_template(
        cls: Type[WF],
        template: WidgetTemplate[T],
        search,
        sized: Callable[[Bounds], Bounds] = Bounds.copy,
        **attrs,
    ) -> Type[WF]:
        t = template.search(search)

        class FormattableWidgetFactory(cls):  # type: ignore
            def format(self) -> List[T]:
                properties = {k: getattr(self, v) for k, v in attrs.items()}
                return [template.set(t, sized(self.bounds), **properties)]

        return FormattableWidgetFactory


@dataclass
class LabelFactory(WidgetFactory[T]):
    text: str


@dataclass
class PVWidgetFactory(WidgetFactory[T]):
    pv: str


@dataclass
class ActionFactory(WidgetFactory[T]):
    label: str
    pv: str
    value: str


class GroupType(Enum):
    GROUP = "GROUP"
    SCREEN = "SCREEN"


@dataclass
class GroupFactory(WidgetFactory[T]):
    title: str
    children: List[WidgetFactory[T]]

    @classmethod
    def from_template(
        cls,
        template: WidgetTemplate[T],
        search: GroupType,
        sized: Callable[[Bounds], Bounds] = Bounds.copy,
        make_widgets: Callable[[Bounds, str], List[WidgetFactory[T]]] = None,
        **attrs,
    ) -> Type[GroupFactory[T]]:
        @dataclass
        class FormattableGroupFactory(GroupFactory[T]):
            def format(self) -> List[T]:
                padding = sized(self.bounds)
                texts: List[T] = []
                made_widgets: List[T] = []

                if search == GroupType.SCREEN:
                    properties = {k: getattr(self, v) for k, v in attrs.items()}
                    texts.append(
                        template.set(template.screen, self.bounds, **properties)
                    )
                    # Make screen title
                    if make_widgets:
                        for widget in make_widgets(self.bounds, self.title):
                            texts += widget.format()
                    for c in self.children:
                        c.bounds.x += padding.x
                        c.bounds.y += padding.y
                        texts += c.format()

                if search == GroupType.GROUP:
                    # Make group object
                    if make_widgets:
                        for widget in make_widgets(self.bounds, self.title):
                            made_widgets += widget.format()
                    texts += template.create_group(made_widgets, self.children, padding)
                return texts

            def __post_init__(self):
                padding = sized(self.bounds)
                self.bounds = replace(self.bounds, w=padding.w, h=padding.h)

        return FormattableGroupFactory


def max_x(widgets: List[WidgetFactory[T]]) -> int:
    """Given multiple widgets, calulate the maximum x position that they occupy"""
    if widgets:
        return max(w.bounds.x + w.bounds.w for w in widgets)
    else:
        return 0


def max_y(widgets: List[WidgetFactory[T]]) -> int:
    """Given multiple widgets, calulate the maximum y position that they occupy"""
    if widgets:
        return max(w.bounds.y + w.bounds.h for w in widgets)
    else:
        return 0


def next_x(widgets: List[WidgetFactory[T]], spacing: int = 0) -> int:
    """Given multiple widgets, calulate the next feasible location for an
    additional widget in the x axis"""
    if widgets:
        return max_x(widgets) + spacing
    else:
        return 0


def next_y(widgets: List[WidgetFactory[T]], spacing: int = 0) -> int:
    """Given multiple widgets, calulate the next feasible location for an
    additional widget in the y axis"""
    if widgets:
        return max_y(widgets) + spacing
    else:
        return 0


@dataclass
class LayoutProperties:
    spacing: Annotated[int, desc("Spacing between widgets")]
    title_height: Annotated[int, desc("Height of screen title bar")]
    max_height: Annotated[int, desc("Max height of the screen")]
    group_label_height: Annotated[int, desc("Height of the group title label")]
    label_width: Annotated[int, desc("Width of the labels describing widgets")]
    widget_width: Annotated[int, desc("Width of the widgets")]
    widget_height: Annotated[int, desc("Height of the widgets (Labels use this too)")]
    group_widget_indent: Annotated[
        int, desc("Indentation of widgets within groups. Defaults to 0")
    ] = 0
    group_width_offset: Annotated[
        int, desc("Additional border width when using group objects. Defaults to 0")
    ] = 0


@dataclass
class ScreenWidgets(Generic[T]):
    label_cls: Type[LabelFactory[T]]
    led_cls: Type[PVWidgetFactory[T]]
    progress_bar_cls: Type[PVWidgetFactory[T]]
    # TODO: add bitfield, progress_bar, plot, table, image
    text_read_cls: Type[PVWidgetFactory[T]]
    check_box_cls: Type[PVWidgetFactory[T]]
    combo_box_cls: Type[PVWidgetFactory[T]]
    text_write_cls: Type[PVWidgetFactory[T]]
    action_button_cls: Type[ActionFactory[T]]

    def pv_widget(
        self,
        widget: Union[ReadWidget, WriteWidget],
        bounds: Bounds,
        pv: str,
        prefix: str,
    ) -> PVWidgetFactory[T]:
        """Convert a component that reads or writes PV's into its WidgetFactory equivalent

        Args:
            widget: The read/write widget property of a component
            bounds: The size and position of the widget (x,y,w,h).
            pv: The process variable assigned to a component

        Returns:
            A WidgetFactory representing the component
        """

        widget_factory: Dict[type, Type[PVWidgetFactory[T]]] = {
            # Currently supported instances of ReadWidget/WriteWidget Components
            LED: self.led_cls,
            ProgressBar: self.progress_bar_cls,
            TextRead: self.text_read_cls,
            CheckBox: self.check_box_cls,
            ComboBox: self.combo_box_cls,
            TextWrite: self.text_write_cls,
        }
        if isinstance(widget, (TextRead, TextWrite)):
            bounds.h *= widget.lines
        return widget_factory[type(widget)](bounds, prefix + pv)


@dataclass
class Screen(Generic[T]):
    screen_cls: Type[GroupFactory[T]]
    group_cls: Type[GroupFactory[T]]
    screen_widgets: ScreenWidgets
    layout: LayoutProperties
    prefix: str
    components: Dict[str, Component] = field(init=False, default_factory=dict)

    def screen(self, components: Tree[Component], title: str) -> WidgetFactory[T]:
        """Make the contents of the screen and determines the layout of widgets

        Args:
            components: A list of components that make up a device
            title: The title of the screen

        Returns:
            A constructed screen object

        """
        full_w = (
            self.layout.label_width + self.layout.widget_width + 2 * self.layout.spacing
        )
        screen_bounds = Bounds(h=self.layout.max_height)
        widget_dims = dict(w=full_w, h=self.layout.widget_height)
        screen_widgets: List[WidgetFactory[T]] = []
        columns: List[Bounds] = [Bounds(**widget_dims)]
        for c in components:
            last_column_bounds = columns[-1]
            next_column_bounds = Bounds(
                x=next_x(screen_widgets, self.layout.spacing),
                y=0,
                **widget_dims,
            )
            if isinstance(c, Group):
                # Create group widget
                # Note: Group adjusts bounds to fit the components
                screen_widgets.extend(
                    self.make_group_widget(
                        c,
                        screen_bounds=screen_bounds,
                        column_bounds=last_column_bounds,
                        next_column_bounds=next_column_bounds,
                    )
                )

            else:
                # Create top level widget - note this will change columns in place
                screen_widgets.extend(
                    self.make_component_widgets(
                        c,
                        column_bounds=last_column_bounds,
                        parent_bounds=screen_bounds,
                        next_column_bounds=next_column_bounds,
                        group_widget_indent=self.layout.group_widget_indent,
                    )
                )

            if next_column_bounds.y != 0:
                columns.append(next_column_bounds)

        screen_bounds.w = max_x(screen_widgets)
        screen_bounds.h = max_y(screen_widgets)
        return self.screen_cls(screen_bounds, title, screen_widgets)

    def component(
        self, c: Component, bounds: Bounds, group_widget_indent: int, add_label=True
    ) -> Iterator[WidgetFactory[T]]:
        """Convert a component into its WidgetFactory equivalents

        Args:
            c: Component object extracted from a producer.yaml
            bounds: The size and position of the component widgets (x,y,w,h).
            add_label: Whether the component has an associated label. Defaults to True.

        Yields:
            A collection of widgets representing the component
        """

        def indent_widget(bounds: Bounds, indentation: int) -> Bounds:
            """Shifts the x position of a widget. Used on top level widgets to align
            them with group indentation"""
            return Bounds(bounds.x + indentation, bounds.y, bounds.w, bounds.h)

        # Widgets are allowed to expand bounds
        if not isinstance(c, SignalRef):
            self.components[c.name] = c

        if add_label:
            left, bounds = bounds.split(self.layout.label_width, self.layout.spacing)
            yield self.screen_widgets.label_cls(
                indent_widget(left, group_widget_indent), c.get_label()
            )

        if isinstance(c, SignalX):
            yield self.screen_widgets.action_button_cls(
                indent_widget(bounds, group_widget_indent),
                c.get_label(),
                c.pv,
                c.value,
            )
        elif isinstance(c, SignalR) and c.widget:
            yield self.screen_widgets.pv_widget(
                c.widget, indent_widget(bounds, group_widget_indent), c.pv, self.prefix
            )
        elif isinstance(c, SignalRW) and c.read_pv and c.read_widget and c.widget:
            left, right = bounds.split(
                int((bounds.w - self.layout.spacing) / 2), self.layout.spacing
            )
            yield self.screen_widgets.pv_widget(
                c.widget, indent_widget(left, group_widget_indent), c.pv, self.prefix
            )
            yield self.screen_widgets.pv_widget(
                c.read_widget,
                indent_widget(right, group_widget_indent),
                c.read_pv,
                self.prefix,
            )
        elif isinstance(c, (SignalW, SignalRW)) and c.widget:
            yield self.screen_widgets.pv_widget(
                c.widget, indent_widget(bounds, group_widget_indent), c.pv, self.prefix
            )
        elif isinstance(c, SignalRef):
            yield from self.component(
                self.components[c.name],
                indent_widget(bounds, group_widget_indent),
                add_label,
            )
        # TODO: Need to handle DeviceRef

    def make_group_widget(
        self,
        c: Group[Component],
        screen_bounds: Bounds,
        column_bounds: Bounds,
        next_column_bounds: Bounds,
    ) -> List[WidgetFactory[T]]:
        """Generate widget from Group component data

        Args:
            c: Group of Component objects extracted from a device.yaml
            screen_bounds: Bounds of the top level screen
            column_bounds: Bounds of widget in current column
            parent_bounds: Size constraints from the object containing the widgets
            next_column_bounds: Bounds of widget in next column should widgets exceed
                height limits

        Returns:
            A collection of widgets representing the component

        Side Effect:
            One of {column_bounds,next_column_bounds}.y will be updated to fit the
            added widget

        """
        group = self.group(
            c,
            bounds=Bounds(column_bounds.x, column_bounds.y, h=screen_bounds.h),
        )

        if group.bounds.h + group.bounds.y <= screen_bounds.h:
            # Group fits in this column
            group.bounds.w += self.layout.group_width_offset

            # Update y for current column
            column_bounds.y = group.bounds.y + group.bounds.h + self.layout.spacing

        else:
            # Won't fit in first column, force it into next column
            group = self.group(
                c,
                bounds=Bounds(
                    next_column_bounds.x, next_column_bounds.y, h=screen_bounds.h
                ),
            )
            group.bounds.w += self.layout.group_width_offset

            next_column_bounds.y = group.bounds.y + group.bounds.h + self.layout.spacing

        return [group]

    def group(self, group: Group[Component], bounds: Bounds) -> WidgetFactory[T]:
        """Convert components within groups into widgets and lay them out within a
        group object.

        Args:
            group: Group of child components in a Layout
            bounds: The size and position of the group widget (x,y,w,h).

        Returns:
            A group object containing child widgets
        """

        full_w = (
            self.layout.label_width + self.layout.widget_width + 2 * self.layout.spacing
        )
        column_bounds = Bounds(w=full_w, h=self.layout.widget_height)
        widgets: List[WidgetFactory[T]] = []
        assert isinstance(group.layout, Grid), "Can only do grid at the moment"
        for c in group.children:
            if isinstance(c, Group):
                # TODO: make a new screen
                raise NotImplementedError(c)
            else:
                next_column_bounds = Bounds(
                    x=next_x(widgets, self.layout.spacing),
                    w=full_w,
                    h=self.layout.widget_height,
                )
                widgets = widgets + self.make_component_widgets(
                    c,
                    column_bounds=column_bounds,
                    parent_bounds=bounds,
                    next_column_bounds=next_column_bounds,
                    add_label=group.layout.labelled,
                )
                if next_column_bounds.y != 0:
                    # We have moved onto the next column
                    column_bounds = next_column_bounds

        bounds.h = max_y(widgets)
        bounds.w = max_x(widgets)
        return self.group_cls(bounds, group.get_label(), widgets)

    def make_component_widgets(
        self,
        c: Component,
        column_bounds: Bounds,
        parent_bounds: Bounds,
        next_column_bounds: Bounds,
        add_label=True,
        group_widget_indent: int = 0,
    ) -> List[WidgetFactory[T]]:
        """Generate widgets from component data and position them in a grid format

        Args:
            c: Component object extracted from a device.yaml
            column_bounds: Bounds of widget in current column
            parent_bounds: Size constraints from the object containing the widgets
            next_column_bounds: Bounds of widget in next column should widgets exceed
                height limits
            add_label: Whether the widget should have an associated label.
                Defaults to True.
            group_widget_indent: The x offset of widgets within groups.

        Returns:
            A collection of widgets representing the component

        Side Effect:
            One of {column_bounds,next_column_bounds}.y will be updated to fit the
            added widget

        """
        widgets = list(self.component(c, column_bounds, group_widget_indent, add_label))
        if max_y(widgets) > parent_bounds.h:
            # Add to next column
            widgets = list(
                self.component(c, next_column_bounds, group_widget_indent, add_label)
            )
            next_column_bounds.y = next_y(widgets, self.layout.spacing)
        else:
            column_bounds.y = next_y(widgets, self.layout.spacing)

        return widgets


def concat(items: List[List[T]]) -> List[T]:
    return [x for seq in items for x in seq]


def split_with_sep(text: str, sep: str, maxsplit: int = -1) -> List[str]:
    return [t + sep for t in text.split(sep, maxsplit=maxsplit)]


def with_title(spacing, title_height: int) -> Callable[[Bounds], Bounds]:
    return Bounds(
        spacing, spacing + title_height, 2 * spacing, 2 * spacing + title_height
    ).added_to


class EdlTemplate(WidgetTemplate[str]):
    def __init__(self, text: str):
        assert "endGroup" not in text, "Can't do groups"
        self.screen, text = split_with_sep(text, "\nendScreenProperties\n", 1)
        self.widgets = split_with_sep(text, "\nendObjectProperties\n")

    def set(self, t: str, bounds: Bounds = None, **properties) -> str:
        if bounds:
            for k in "xywh":
                properties[k] = getattr(bounds, k)
        for item, value in properties.items():
            multiline = re.compile(r"^%s {[^}]*}$" % item, re.MULTILINE | re.DOTALL)
            if multiline.search(t):
                pattern = multiline
                lines = str(value).splitlines()
                value = "\n".join(["{"] + [f'  "{x}"' for x in lines] + ["}"])
            else:
                # Single line
                pattern = re.compile(r"^%s .*$" % item, re.MULTILINE)
                if isinstance(value, str):
                    value = f'"{value}"'
            t, n = pattern.subn(f"{item} {value}", t)
            assert n == 1, f"No replacements made for {item}"
        return t

    def search(self, search: str) -> str:
        matches = [t for t in self.widgets if re.search(search, t)]
        assert len(matches) == 1, f"Got {len(matches)} matches for {search!r}"
        return matches[0]

    def create_group(
        self,
        group_object: List[str],
        children: List[WidgetFactory[str]],
        padding: Bounds = Bounds(),
    ) -> List[str]:

        texts: List[str] = []

        for c in children:
            c.bounds.x += padding.x
            c.bounds.y += padding.y
            texts += c.format()

        return group_object + texts


class AdlTemplate(WidgetTemplate[str]):
    def __init__(self, text: str):
        assert "children {" not in text, "Can't do groups"
        widgets = split_with_sep(text, "\n}\n")
        self.screen = "".join(widgets[:3])
        self.widgets = widgets[3:]

    def set(self, t: str, bounds: Bounds = None, **properties) -> str:
        if bounds:
            properties["x"] = bounds.x
            properties["y"] = bounds.y
            properties["width"] = bounds.w
            properties["height"] = bounds.h
        for item, value in properties.items():
            # Only need single line
            pattern = re.compile(r"^(\s*%s)=.*$" % item, re.MULTILINE)
            if isinstance(value, str):
                value = f'"{value}"'
            t, n = pattern.subn(r"\g<1>=" + str(value), t)
            assert n == 1, f"No replacements made for {item}"
        return t

    def search(self, search: str) -> str:
        matches = [t for t in self.widgets if re.search(search, t)]
        assert len(matches) == 1, f"Got {len(matches)} matches for {search!r}"
        return matches[0]

    def create_group(
        self,
        group_object: List[str],
        children: List[WidgetFactory[str]],
        padding: Bounds = Bounds(),
    ) -> List[str]:

        texts: List[str] = []

        for c in children:
            c.bounds.x += padding.x
            c.bounds.y += padding.y
            texts += c.format()

        return group_object + texts


class BobTemplate(WidgetTemplate[etree.ElementBase]):
    """Extract and modify elements from a template .bob file."""

    def __init__(self, text: str):
        """Parse an XML string to an element tree object."""

        self.tree = etree.parse(text)
        self.screen = self.search("Display")

    def set(
        self, t: etree.ElementBase, bounds: Bounds = None, **properties
    ) -> etree.ElementBase:
        """Modify template elements (widgets) with component data.

        Args:
            t: A template element.
            bounds: The size and position of the widget.
            **properties: The element properties (SubElements) to update.
                In the form: {[SubElement]: [Value]}

        Returns:
            The modified element.
        """
        t_copy = deepcopy(t)
        if bounds:
            properties["x"] = bounds.x
            properties["y"] = bounds.y
            properties["width"] = bounds.w
            properties["height"] = bounds.h
        for item, value in properties.items():
            try:
                t_copy.xpath(f"./{item}")[0].text = str(value)
            except IndexError as idx:
                name = t_copy.find("name").text
                raise ValueError(f"Failed to locate '{item}' in {name}") from idx
        return t_copy

    def search(self, search: str) -> etree.ElementBase:
        """Locate and extract elements from the Element tree.

        Args:
            search: The unique name of the element to extract.
                Can be found in its name subelement.

        Returns:
            The extracted element.
        """

        tree_copy = deepcopy(self.tree)
        # 'name' is the unique ID for each element
        matches = [
            element.getparent()
            for element in tree_copy.iter("name")
            if element.text == search
        ]
        assert len(matches) == 1, f"Got {len(matches)} matches for {search!r}"

        # Isolate the screen properties
        if matches[0].tag == "display":
            for child in matches[0]:
                if child.tag == "widget":
                    matches[0].remove(child)

        return matches[0]

    def create_group(
        self,
        group_object: List[etree.ElementBase],
        children: List[WidgetFactory[etree.ElementBase]],
        padding: Bounds = Bounds(),
    ) -> List[etree.ElementBase]:
        """Create an xml group object from a list of child widgets

        Args:
            group_object: Templated group xml element
            children: List of child widgets within the group
            padding: Additional placement data to fit the children to the group object.
                Defaults to Bounds().

        Returns:
            An xml group with children attached as subelements
        """
        assert (
            len(group_object) == 1
        ), f"Size of group_object is {len(group_object)}, should be 1"
        for c in children:
            group_object[0].append(c.format()[0])
        return group_object



from typing import Type, TypeVar


def truncate_description(desc: str) -> str:
    """Take the first line of a multiline description, truncated to 40 chars"""
    first_line = desc.strip().split("\n")[0]
    return first_line[:40]


def join_lines(lines, indent=0):
    return ("\n" + (" " * indent)).join(lines)


def get_param_set(driver: str) -> str:
    return "asynParamSet" if driver == "asynPortDriver" else driver + "ParamSet"


T = TypeVar("T")


def narrow_type(obj, type: Type[T]) -> T:
    """Narrow the type of `obj` to the given type to pass mypy checks

    Raises `AssertionError` if `obj` is not an instance of `type`

    """
    assert isinstance(obj, type), "%s is not an instance of %s" % (
        obj,
        type.__name__,
    )
    return obj



import re
from pathlib import Path
from typing import Dict, List, Tuple

from pvi._produce.asyn import AsynParameter, AsynProducer
from pvi.device import Grid, Group

from ._asyn_convert import (
    Action,
    AsynRecord,
    Parameter,
    Readback,
    RecordError,
    SettingPair,
)

OVERRIDE_DESC = "# Overriding value in auto-generated template"


class TemplateConverter:
    def __init__(self, templates: List[Path]):
        self.templates = templates
        self._text = [t.read_text() for t in self.templates]

    def top_level_text(self, driver_name: str):
        extracted_templates = []
        for text in self._text:
            record_extractor = RecordExtractor(text)
            extracted_templates.append(record_extractor.get_top_level_text(driver_name))
        return extracted_templates

    def convert(self) -> AsynProducer:
        def get_prefix(texts: List[str]) -> str:
            # e.g. from: record(waveform, "$(P)$(R)FilePath")
            # extract: $(P)$(R)
            prefix_extractor = re.compile(
                r'(?:record\()(?:[^,]*)(?:[^"]*)(?:")((?:\$\([^)]\))*)(?:[^"]*)'
            )
            for text in texts:
                prefixes = re.findall(prefix_extractor, text)
            prefixes = list(set(prefixes))
            if len(prefixes) > 1:
                raise ValueError("Not all asyn records have the same macro prefix")
            return prefixes.pop()

        def get_asyn_parameters(texts: List[str]) -> Dict[int, str]:
            # e.g. from: field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
            # extract: $(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH
            asyn_parameter_extractor = r'(?:@asyn\()([^"]*)'
            for text in texts:
                asyn_parameters = re.findall(asyn_parameter_extractor, text)
            # then: remove final close bracket and driver param name
            # $(PORT),$(ADDR=0),$(TIMEOUT=1)
            asyn_parameters = [match[: match.rfind(")")] for match in asyn_parameters]
            assert asyn_parameters, "No Asyn parameters found"
            if len(set(asyn_parameters)) > 1:
                print(
                    "More than one set of asyn params found. Taking the first instance"
                )
            return {i: p.strip() for i, p in enumerate(asyn_parameters[0].split(","))}

        asyn_vars = get_asyn_parameters(self._text)

        return AsynProducer(
            prefix=get_prefix(self._text),
            label=self.templates[0].stem,
            asyn_port=asyn_vars.get(0, "$(PORT)"),
            address=asyn_vars.get(1, "$(ADDR=0)"),
            timeout=asyn_vars.get(2, "$(TIMEOUT=1)"),
            parent="asynPortDriver",
            parameters=[
                Group(
                    name="ComponentGroupOne",
                    layout=Grid(),
                    children=self._extract_components(),
                )
            ],
        )

    def _extract_components(self) -> List[AsynParameter]:
        components = []
        for text in self._text:
            record_extractor = RecordExtractor(text)
            asyn_records = record_extractor.get_asyn_records()
            for parameter in RecordRoleSorter.sort_records(asyn_records):
                component = parameter.generate_component()
                components.append(component)
        return components


class RecordExtractor:
    def __init__(self, text):
        self._text = text

    def _extract_record_strs(self):
        # extract a whole record definition inc. fields and leading empty lines e.g.
        # record(waveform, "$(P)$(R)FilePath")
        # {
        #    field(PINI, "YES")
        #    field(DTYP, "asynOctetWrite")
        #    field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
        #    field(FTVL, "CHAR")
        #    field(NELM, "256")
        #    info(autosaveFields, "VAL")
        # }
        record_extractor = re.compile(r"\s*^[^#\n]*record\([^{]*{[^}]*}", re.MULTILINE)
        return re.findall(record_extractor, self._text)

    def _parse_record(self, record_str: str) -> Tuple:
        # extract three groups from a record definition e.g.
        # from:
        # record(waveform, "$(P)$(R)FilePath")
        # {
        #    #field(PINI, "YES")
        #    field(DTYP, "asynOctetWrite")
        #    field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
        #    field(FTVL, "CHAR")
        #    field(NELM, "256")
        #    info(autosaveFields, "VAL")
        # }
        # extract:
        # Group 1 - record type: waveform
        # Group 2 - record name exc. prefix: FilePath
        # Group 3 - all fields:
        #    #field(PINI, "YES")
        #    field(DTYP, "asynOctetWrite")
        #    field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
        #    field(FTVL, "CHAR")
        #    field(NELM, "256")
        #    info(autosaveFields, "VAL")
        #
        record_parser = re.compile(
            r'(?:record\()([^,]*)(?:[^"]*)(?:")'
            r'(?:(?:\$\([a-zA-Z0-9]\))*)([^"]*)'
            r'(?:")(?:[^{]*)(?:{)([^}]*)(?:})'
        )
        return re.findall(record_parser, record_str)[0]

    def _extract_fields(self, fields_str: str) -> List[Tuple[str, str]]:
        # extract two groups from a field e.g.
        # from: field(PINI, "YES")
        # extract:
        # Group 1 - Field: PINI
        # Group 2 - Value: YES
        field_extractor = re.compile(
            r'^[^#\n]*(?:field\()([^,]*)(?:,)(?:[^"]*)(?:")([^"]*)(?:")', re.MULTILINE
        )
        return re.findall(field_extractor, fields_str)

    def _extract_infos(self, fields_str: str) -> List[Tuple[str, str]]:
        # extract two groups from an info tag e.g.
        # from: info(autosaveFields, "VAL")
        # extract:
        # Group 1 - Field: autosaveFields
        # Group 2 - Value: VAL
        info_extractor = re.compile(
            r'^[^#\n]*(?:info\()([^,]*)(?:,)(?:[^"]*)(?:")([^"]*)(?:")', re.MULTILINE
        )
        return re.findall(info_extractor, fields_str)

    def _create_asyn_record(self, record_str: str) -> AsynRecord:
        record_type, record_name, record_fields = self._parse_record(record_str)
        fields = dict(self._extract_fields(record_fields))
        info = dict(self._extract_infos(record_fields))
        record = AsynRecord(
            name=record_name, type=record_type, fields=fields, infos=info
        )
        return record

    def get_asyn_records(self) -> List[AsynRecord]:
        record_strs = self._extract_record_strs()
        record_list = []
        for record_str in record_strs:
            try:
                record_list.append(self._create_asyn_record(record_str))
            except RecordError:
                pass
        return record_list

    def _create_stream_record(self, record_str):
        raise RecordError

    def get_stream_records(self):
        return []

    def get_top_level_text(self, driver_name: str) -> str:
        record_strs = self._extract_record_strs()
        top_level_str = self._text
        for record_str in record_strs:
            try:
                self._create_asyn_record(record_str)
            except RecordError:
                try:
                    self._create_stream_record(record_str)
                except RecordError:
                    pass
            else:
                top_level_str = top_level_str.replace(record_str, "")

        # Get override strings for setting pair clashes
        asyn_records = self.get_asyn_records()

        setting_pairs = [
            p
            for p in RecordRoleSorter.sort_records(asyn_records)
            if isinstance(p, SettingPair)
        ]
        overrides = [
            setting_pair.get_naming_overrides()
            for setting_pair in setting_pairs
            if setting_pair.has_clashes()
        ]
        record_lines = {
            self._parse_record(record_str)[1]: record_str.splitlines()
            for record_str in record_strs
        }

        def keep_line(line: str, clashing_fields: List[str]) -> bool:
            extracted_field = self._extract_fields(line) or self._extract_infos(line)
            return not extracted_field or (extracted_field[0][0] in clashing_fields)

        override = [
            "\n".join(
                [OVERRIDE_DESC]
                + [
                    line
                    for line in record_lines[record_name]
                    if line and keep_line(line, clashing_fields)
                ]
            )
            for record_name, clashing_fields in overrides
        ]

        top_level_str = self._add_param_template_include(top_level_str, driver_name)
        top_level_str += "\n\n".join(override)
        return top_level_str

    def _add_param_template_include(self, top_level_str: str, driver_name: str) -> str:
        top_level_str = f'include "{driver_name}Parameters.template"\n' + top_level_str
        return top_level_str


class RecordRoleSorter:
    @staticmethod
    def sort_records(records: List[AsynRecord]) -> List[Parameter]:
        def _sort_inputs_outputs(
            records: List[AsynRecord],
        ) -> Tuple[List[AsynRecord], List[AsynRecord]]:
            inp_records = [r for r in records if "INP" in r.fields]
            write_records = [r for r in records if "OUT" in r.fields]

            # Move waveform records with asynOctetWrite from read to write
            read_records = []
            for r in inp_records:
                if r.type == "waveform" and (
                    r.fields["DTYP"] == "asynOctetWrite"
                    or r.fields["DTYP"].endswith("ArrayOut")
                ):
                    write_records.append(r)
                else:
                    read_records.append(r)
            return read_records, write_records

        read_records, write_records = _sort_inputs_outputs(records)
        parameters: List[Parameter] = []
        parameters += ParameterRoleMatcher.get_actions(read_records, write_records)
        parameters += ParameterRoleMatcher.get_readbacks(read_records, write_records)
        parameters += ParameterRoleMatcher.get_setting_pairs(
            read_records, write_records
        )
        return parameters


class ParameterRoleMatcher:
    @staticmethod
    def get_actions(
        read_records: List[AsynRecord], write_records: List[AsynRecord]
    ) -> List[Action]:
        actions = [
            Action(write_record=w)
            for w in write_records
            if w.get_parameter_name()
            not in [r.get_parameter_name() for r in read_records]
        ]
        return actions

    @staticmethod
    def get_readbacks(
        read_records: List[AsynRecord], write_records: List[AsynRecord]
    ) -> List[Readback]:
        readbacks = [
            Readback(read_record=r)
            for r in read_records
            if r.get_parameter_name()
            not in [w.get_parameter_name() for w in write_records]
        ]
        return readbacks

    @staticmethod
    def get_setting_pairs(
        read_records: List[AsynRecord], write_records: List[AsynRecord]
    ) -> List[SettingPair]:
        setting_pairs = [
            SettingPair(read_record=r, write_record=w)
            for r in read_records
            for w in write_records
            if r.get_parameter_name() == w.get_parameter_name()
        ]
        return setting_pairs



import json
import os
from pathlib import Path
from typing import List, Optional, Type

import typer

from pvi import __version__
from pvi._convert._source_convert import SourceConverter
from pvi._convert._template_convert import TemplateConverter
from pvi._format import Formatter
from pvi._produce import Producer
from pvi._produce.asyn import AsynParameter, AsynProducer
from pvi._pv_group import group_parameters
from pvi._schema_utils import make_json_schema
from pvi._yaml_utils import deserialize_yaml, serialize_yaml
from pvi.device import Device, walk

app = typer.Typer()
convert_app = typer.Typer()
app.add_typer(convert_app, name="convert", help="Convert a module to use PVI")


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print the version and exit",
    )
):
    """PVI builder interface"""


@app.command()
def schema(output: Path = typer.Argument(..., help="filename to write the schema to")):
    """Write the JSON schema for the pvi interface or producer interface"""
    cls: Type
    assert output.name.endswith(
        ".schema.json"
    ), f"Expected '{output.name}' to end with '.schema.json'"
    if output.name == "pvi.device.schema.json":
        cls = Device
    elif output.name == "pvi.producer.schema.json":
        cls = Producer
    elif output.name == "pvi.formatter.schema.json":
        cls = Formatter
    else:
        typer.echo(f"Don't know how to create {output.name}")
        raise typer.Exit(code=1)
    schema = make_json_schema(cls)
    output.write_text(json.dumps(schema, indent=2))


@app.command()
def produce(
    output: Path = typer.Argument(..., help="filename to write the product to"),
    producer: Path = typer.Argument(..., help="path to the producer .pvi.yaml file"),
    yaml_paths: Optional[List[Path]] = typer.Option(
        None, help="Paths to directories with .pvi.producer.yaml files"
    ),
):
    """Create template/csv/device/other product from producer YAML"""
    if yaml_paths is None:
        yaml_paths = []

    producer_inst = deserialize_yaml(Producer, producer)
    if output.suffix == ".template":
        producer_inst.produce_records(output)
    elif output.suffix == ".csv":
        producer_inst.produce_csv(output)
    elif output.suffixes == [".pvi", ".device", ".yaml"]:
        device = producer_inst.produce_device()
        serialize_yaml(device, output)
    else:
        producer_inst.produce_other(output, yaml_paths)


@app.command()
def format(
    output: Path = typer.Argument(..., help="Directory to write output file(s) to"),
    producer: Path = typer.Argument(..., help="path to the .pvi.producer.yaml file"),
    formatter: Path = typer.Argument(..., help="path to the .pvi.formatter.yaml file"),
    yaml_paths: Optional[List[Path]] = typer.Option(
        None, help="Paths to directories with .pvi.producer.yaml files"
    ),
):
    """Create screen product from producer and formatter YAML"""
    if yaml_paths is None:
        yaml_paths = []

    producer_inst: Producer = deserialize_yaml(Producer, producer)
    producer_inst.deserialize_parents(yaml_paths)
    device = producer_inst.produce_device()

    formatter_inst: Formatter = deserialize_yaml(Formatter, formatter)
    formatter_inst.format(device, producer_inst.prefix, output)


@convert_app.command()
def asyn(
    output: Path = typer.Argument(..., help="Directory to write output file(s) to"),
    cpp: Path = typer.Argument(..., help="Path to the .cpp file to convert"),
    h: Path = typer.Argument(..., help="Path to the .h file to convert"),
    templates: Optional[List[Path]] = typer.Argument(
        None, help="Paths to .template files to convert"
    ),
    yaml_paths: Optional[List[Path]] = typer.Option(
        None, help="Paths to directories with .pvi.producer.yaml files"
    ),
):
    """Convert cpp/h/template to producer YAML and stripped cpp/h/template"""
    if yaml_paths is None:
        yaml_paths = []

    if not output.exists():
        os.mkdir(output)

    drv_infos = []
    producer = AsynProducer(
        prefix="$(P)$(R)",
        label=h.stem,
        asyn_port="$(PORT)",
        address="$(ADDR=0)",
        timeout="$(TIMEOUT=1)",
        parent="asynPortDriver",
        parameters=[],
    )

    # if a template is given, convert it and populate drv_infos
    if templates:
        template_converter = TemplateConverter(templates)

        # Generate initial yaml to provide parameter info strings to source converter
        producer = template_converter.convert()
        drv_infos = [
            parameter.get_drv_info()
            for parameter in walk(producer.parameters)
            if isinstance(parameter, AsynParameter)
        ]

    source_converter = SourceConverter(cpp, h, yaml_paths, drv_infos)

    if templates:
        # Process and recreate template files - pass source device for param set include
        extracted_templates = template_converter.top_level_text(
            source_converter.device_class
        )
        for template_text, template_path in zip(extracted_templates, templates):
            (output / template_path.name).write_text(template_text)

    # Process and recreate source files
    extracted_source = source_converter.get_top_level_text()
    (output / cpp.name).write_text(extracted_source.cpp)
    (output / h.name).write_text(extracted_source.h)

    # Update yaml based on source file definitions and write
    producer.parent = source_converter.parent_class
    if source_converter.define_strings:
        index_map = source_converter.get_info_index_map()
        for parameter in walk(producer.parameters):
            if isinstance(parameter, AsynParameter):
                parameter.index_name = index_map.get(
                    parameter.get_drv_info(), "INPUT MANUALLY"
                )

    serialize_yaml(
        producer, output / f"{source_converter.device_class}.pvi.producer.yaml"
    )


@app.command()
def convertplaceholder(
    output: Path = typer.Argument(..., help="Directory to write output file(s) to"),
    cpp: Path = typer.Argument(..., help="Path to the .cpp file to convert"),
    h: Path = typer.Argument(..., help="Path to the .h file to convert"),
    yaml_paths: Optional[List[Path]] = typer.Option(
        None, help="Paths to directories with .pvi.producer.yaml files"
    ),
):
    """Alter cpp and h files of unconverted drivers"""
    if yaml_paths is None:
        yaml_paths = []

    if not output.exists():
        os.mkdir(output)

    drv_infos: List[str] = []
    source_converter = SourceConverter(cpp, h, yaml_paths, drv_infos)
    extracted_source = source_converter.get_top_level_placeholder()

    (output / cpp.name).write_text(extracted_source.cpp)
    (output / h.name).write_text(extracted_source.h)


@app.command()
def regroup(
    output: Path = typer.Argument(..., help="Directory to write output file(s) to"),
    producer_path: Path = typer.Argument(
        ..., help="Path to the producer.yaml file to regroup"
    ),
    ui_paths: List[Path] = typer.Argument(
        ..., help="Paths to the ui files to regroup the PVs by"
    ),
):
    """Regroup a producer.yaml file based on ui files that the PVs appear in"""
    producer = deserialize_yaml(AsynProducer, producer_path)
    producer.parameters = group_parameters(producer, ui_paths)

    # Create new yaml
    if not output.exists():
        os.mkdir(output)
    serialize_yaml(producer, output.joinpath(f"{producer_path.stem}.yaml"))


# test with: pipenv run python -m pvi
if __name__ == "__main__":
    app()



from dataclasses import dataclass
from pathlib import Path
from typing import List

from lxml import etree
from typing_extensions import Annotated

from pvi._schema_utils import desc
from pvi.device import Device

from .base import Formatter
from .utils import (
    ActionFactory,
    BobTemplate,
    Bounds,
    EdlTemplate,
    GroupFactory,
    GroupType,
    LabelFactory,
    LayoutProperties,
    PVWidgetFactory,
    Screen,
    ScreenWidgets,
    WidgetFactory,
    with_title,
)


@dataclass
class DLSFormatter(Formatter):
    spacing: Annotated[int, desc("Spacing between widgets")] = 5
    title_height: Annotated[int, desc("Height of screen title bar")] = 25
    max_height: Annotated[int, desc("Max height of the screen")] = 900
    label_width: Annotated[int, desc("Width of the widget description labels")] = 115
    widget_width: Annotated[int, desc("Width of the widgets")] = 120
    widget_height: Annotated[int, desc("Height of the widgets")] = 20

    def format(self, device: Device, prefix: str, path: Path):
        if path.suffix == ".edl":
            f = self.format_edl
        elif path.suffix == ".bob":
            f = self.format_bob
        else:
            raise ValueError("Can only write .edl or .bob files")
        f(device, prefix, path)

    def format_edl(self, device: Device, prefix: str, path: Path):
        template = EdlTemplate((Path(__file__).parent / "dls.edl").read_text())
        layout_properties = LayoutProperties(
            spacing=self.spacing,
            title_height=self.title_height,
            max_height=self.max_height,
            group_label_height=10,
            label_width=self.label_width,
            widget_width=self.widget_width,
            widget_height=self.widget_height,
            group_widget_indent=5,
            group_width_offset=0,
        )
        screen_widgets = ScreenWidgets(
            label_cls=LabelFactory.from_template(
                template, search='"Label"', value="text"
            ),
            led_cls=PVWidgetFactory.from_template(
                template, search='"LED"', sized=Bounds.square, controlPv="pv"
            ),
            progress_bar_cls=PVWidgetFactory.from_template(
                template, search='"ProgressBar"', indicatorPv="pv"
            ),
            text_read_cls=PVWidgetFactory.from_template(
                template, search='"TextRead"', controlPv="pv"
            ),
            check_box_cls=PVWidgetFactory.from_template(
                template, search='"CheckBox"', controlPv="pv"
            ),
            combo_box_cls=PVWidgetFactory.from_template(
                template, search='"ComboBox"', controlPv="pv"
            ),
            text_write_cls=PVWidgetFactory.from_template(
                template, search='"TextWrite"', controlPv="pv"
            ),
            action_button_cls=ActionFactory.from_template(
                template,
                search='"SignalX"',
                onLabel="label",
                offLabel="label",
                controlPv="pv",
            ),
        )
        screen_title_cls = LabelFactory.from_template(
            template, search='"Title"', value="text"
        )
        group_title_cls = LabelFactory.from_template(
            template, search='"  Group  "', value="text"
        )
        group_box_cls = WidgetFactory.from_template(
            template, search="fillColor index 5"
        )

        def make_group_box(bounds: Bounds, title: str) -> List[WidgetFactory[str]]:
            x, y, w, h = bounds.x, bounds.y, bounds.w, bounds.h
            return [
                group_box_cls(
                    Bounds(
                        x,
                        y + layout_properties.spacing,
                        w,
                        h - layout_properties.spacing,
                    )
                ),
                group_title_cls(
                    Bounds(x, y, w, layout_properties.group_label_height),
                    f"  {title}  ",
                ),
            ]

        def make_screen_title(bounds: Bounds, title: str) -> List[WidgetFactory[str]]:
            return [
                screen_title_cls(
                    Bounds(0, 0, bounds.w, layout_properties.title_height), title
                )
            ]

        screen = Screen(
            screen_cls=GroupFactory.from_template(
                template,
                search=GroupType.SCREEN,
                sized=with_title(
                    layout_properties.spacing, layout_properties.title_height
                ),
                make_widgets=make_screen_title,
            ),
            group_cls=GroupFactory.from_template(
                template,
                search=GroupType.GROUP,
                sized=with_title(
                    layout_properties.spacing, layout_properties.group_label_height
                ),
                make_widgets=make_group_box,
            ),
            screen_widgets=screen_widgets,
            prefix=prefix,
            layout=layout_properties,
        )
        title = f"{device.label} - {prefix}"
        texts = screen.screen(device.children, title).format()
        path.write_text("".join(texts))

    def format_bob(self, device: Device, prefix: str, path: Path):
        template = BobTemplate(str(Path(__file__).parent / "dls.bob"))
        # LP DOCS REF: Define the layout properties
        layout_properties = LayoutProperties(
            spacing=self.spacing,
            title_height=self.title_height,
            max_height=self.max_height,
            group_label_height=26,
            label_width=self.label_width,
            widget_width=self.widget_width,
            widget_height=self.widget_height,
            group_widget_indent=18,
            group_width_offset=26,
        )
        # SW DOCS REF: Extract widget types from template file
        screen_widgets = ScreenWidgets(
            label_cls=LabelFactory.from_template(template, search="Label", text="text"),
            led_cls=PVWidgetFactory.from_template(
                template, search="LED", sized=Bounds.square, pv_name="pv"
            ),
            progress_bar_cls=PVWidgetFactory.from_template(
                template, search="ProgressBar", pv_name="pv"
            ),
            text_read_cls=PVWidgetFactory.from_template(
                template, search="TextUpdate", pv_name="pv"
            ),
            check_box_cls=PVWidgetFactory.from_template(
                template, search="ChoiceButton", pv_name="pv"
            ),
            combo_box_cls=PVWidgetFactory.from_template(
                template, search="ComboBox", pv_name="pv"
            ),
            text_write_cls=PVWidgetFactory.from_template(
                template, search="TextEntry", pv_name="pv"
            ),
            action_button_cls=ActionFactory.from_template(
                template, search="ActionButton", text="label", pv_name="pv"
            ),
        )
        # MAKE_WIDGETS DOCS REF: Define screen and group widgets
        screen_title_cls = LabelFactory.from_template(
            template, search="Title", text="text"
        )
        group_object_cls = LabelFactory.from_template(
            template, search="Group", name="text"
        )

        def make_group_object(bounds: Bounds, title: str) -> List[WidgetFactory[str]]:
            return [
                group_object_cls(
                    Bounds(bounds.x, bounds.y, bounds.w, bounds.h), f"{title}"
                )
            ]

        def make_screen_title(bounds: Bounds, title: str) -> List[WidgetFactory[str]]:
            return [
                screen_title_cls(
                    Bounds(0, 0, bounds.w, layout_properties.title_height), title
                )
            ]

        # SCREEN_INI DOCS REF: Construct a screen object
        screen = Screen(
            screen_cls=GroupFactory.from_template(
                template,
                search=GroupType.SCREEN,
                sized=with_title(
                    layout_properties.spacing, layout_properties.title_height
                ),
                make_widgets=make_screen_title,
            ),
            group_cls=GroupFactory.from_template(
                template,
                search=GroupType.GROUP,
                sized=with_title(
                    layout_properties.spacing, layout_properties.group_label_height
                ),
                make_widgets=make_group_object,
            ),
            screen_widgets=screen_widgets,
            prefix=prefix,
            layout=layout_properties,
        )
        # SCREEN_FORMAT DOCS REF: Format the screen
        title = f"{device.label} - {prefix}"
        texts = screen.screen(device.children, title).format()
        # SCREEN_WRITE DOCS REF: Generate the screen file
        # The root:'Display' is always the first element in texts
        ET = etree.fromstring(etree.tostring(texts[0]))
        for element in texts[:0:-1]:
            ET.insert(ET.index(ET.find("height")) + 1, element)
        ET = ET.getroottree()
        ET.write(str(path), pretty_print=True)



import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Tuple

from pvi._produce.asyn import AsynParameter, find_components
from pvi.device import walk


@dataclass
class Source:
    cpp: str
    h: str


class SourceConverter:
    def __init__(
        self, cpp: Path, h: Path, yaml_paths: List[Path], drv_infos: List[str]
    ):
        self.source = Source(cpp.read_text(), h.read_text())
        self.yaml_paths = yaml_paths
        self.parameter_infos = drv_infos

        if cpp.name.startswith("asynNDArrayDriver"):
            # Special case to include NDArrayData to be removed from source files
            # There is also a special case to add it to the paramSet without it
            # appearing in the producer.yaml
            self.parameter_infos.append("ARRAY_DATA")

        self.device_class, self.parent_class = self._extract_device_and_parent_class()
        self.define_strings = self._extract_define_strs(self.parameter_infos)
        self.string_info_map = self._get_string_info_map(self.define_strings)
        self.create_param_strings = self._extract_create_param_strs(
            list(self.string_info_map)
        )
        self.string_index_map = self._get_string_index_map(self.create_param_strings)
        self.declaration_strings = self._extract_index_declarations(
            list(self.string_index_map.values())
        )

    def _extract_device_and_parent_class(self) -> Tuple[str, str]:
        # e.g. extract 'NDPluginDriver' and 'asynNDArrayDriver' from
        # class epicsShareClass NDPluginDriver : public asynNDArrayDriver, public epicsThreadRunable {  # noqa
        class_extractor = re.compile(r"class.* (\w+) : \w+ (\w+).*")
        match = re.search(class_extractor, self.source.h)
        assert match, "Can't find classes"
        classname, parent = match.groups()
        return classname, parent

    def _extract_define_strs(self, info_strings: List[str]) -> List[str]:
        # e.g. extract: #define SimGainXString                "SIM_GAIN_X";
        define_extractor = re.compile(r'\#define[_A-Za-z0-9 ]*"[^"]*".*')
        definitions = re.findall(define_extractor, self.source.h)
        # We only want to extract the defines for the given parameter infos
        definitions = filter_strings(definitions, info_strings)
        return definitions

    def _extract_create_param_strs(self, param_strings: List[str]) -> List[str]:
        # e.g. extract: createParam(SimGainXString, asynParamFloat64, &SimGainX);
        create_param_extractor = re.compile(r"((?:this->)?createParam\([^\)]*\);.*)")
        create_param_strs = re.findall(create_param_extractor, self.source.cpp)
        # We only want to extract the createParam calls for the given parameter strings
        create_param_strs = filter_strings(create_param_strs, param_strings)
        return create_param_strs

    def _extract_index_declarations(self, index_names: List[str]) -> List[str]:
        # e.g. extract:     int SimGainX;
        declaration_extractor = re.compile(r"\s*int [^;]*;")
        declarations = re.findall(declaration_extractor, self.source.h)
        # We only want to extract the declarations for the given index names - this also
        # filters any generic int parameter definitions in the class and some comments
        declarations = filter_strings(declarations, index_names)
        return declarations

    def _get_string_info_map(self, define_strings: List[str]) -> Dict[str, str]:
        string_info_map = dict(
            self._parse_definition_str(definition) for definition in define_strings
        )
        return string_info_map

    @staticmethod
    def _parse_definition_str(definition_str: str) -> Tuple[str, str]:
        # e.g. from: #define SimGainXString                "SIM_GAIN_X";
        # extract:
        # Group1: SimGainXString
        # Group2: SIM_GAIN_X
        define_extractor = re.compile(r'(?:\#define) (\w+) *"([^"]*)')
        string_info_pair = re.findall(define_extractor, definition_str)[0]
        return string_info_pair

    def _get_string_index_map(self, create_param_strings: List[str]) -> Dict[str, str]:
        string_index_dict = dict(
            [
                self._parse_create_param_str(create_param_str)
                for create_param_str in create_param_strings
            ]
        )
        return string_index_dict

    @staticmethod
    def _parse_create_param_str(create_param_str: str) -> Tuple[str, str]:
        # e.g. from: createParam(SimGainXString, asynParamFloat64, &SimGainX);
        # extract: SimGainXString,               asynParamFloat64, &SimGainX
        create_param_extractor = re.compile(r"(?:createParam\()([^\)]*)(?:\))")
        # There are two signatures for createParam. 3 argument and 4 argument
        # I think we are only ever interested in the final three arguments
        create_param_args = (
            re.findall(create_param_extractor, create_param_str)[0]
            .replace(" ", "")
            .split(",")[-3:]
        )
        string_index_pair = (create_param_args[0], create_param_args[2].strip("&"))
        return string_index_pair

    def get_info_index_map(self) -> Dict[str, str]:
        index_string_map = dict(
            (index, string) for string, index in self.string_index_map.items()
        )
        info_index_map = dict(
            (self.string_info_map[string], index)
            for index, string in index_string_map.items()
        )
        return info_index_map

    def get_top_level_text(self) -> Source:
        extracted_cpp = self._convert_cpp(self.source.cpp)
        extracted_h = self._convert_h(self.source.h)

        return Source(extracted_cpp, extracted_h)

    def get_top_level_placeholder(self) -> Source:
        extracted_cpp = self._convert_cpp_placeholder(self.source.cpp)
        extracted_h = self._convert_h_placeholder(self.source.h)

        return Source(extracted_cpp, extracted_h)

    def _convert_cpp(self, original_cpp_text: str) -> str:
        unwanted_strings = self.create_param_strings

        # Re-create text with unwanted lines ommited
        lines = original_cpp_text.splitlines()
        lines = [
            line
            for line in lines
            if line.lstrip()
            not in [unwanted_string.lstrip() for unwanted_string in unwanted_strings]
        ]
        cpp_text = "\n".join(lines)

        # Insert accessors for parameters that have been moved to the param set
        parameters = list(self.string_index_map.values())
        parent_components = find_components(self.parent_class, self.yaml_paths)
        parameters += [
            parameter.get_index_name()
            for parameter in walk(parent_components)
            if isinstance(parameter, AsynParameter)
        ]
        cpp_text = self._insert_param_set_accessors(cpp_text, parameters)

        # Add the param set parameter to the constructor declaration
        cpp_text = cpp_text.replace(
            f"::{self.device_class}(",
            f"::{self.device_class}({self.device_class}ParamSet* paramSet, ",
        )

        # Add the param set parameter to the constructor call in the extern "C" function
        driver = self.device_class

        # Create paramSet and insert it into constructor arguments
        # Group 1 is leading whitespace - used to indent added line
        # Group 2 is the rest of the line up to driver constructor opening bracket
        cpp_text = re.sub(
            r"^(\s*)(.*new " + driver + r"\()",
            r"\1"
            + f"{driver}ParamSet* paramSet = new {driver}ParamSet;\n"
            + r"\1\2paramSet, ",
            cpp_text,
            flags=re.MULTILINE,  # Make ^ match start of line
        )

        # Add the initialiser list base class param set parameter
        parent_param_set = get_param_set(self.parent_class)
        cpp_text = cpp_text.replace(
            f"{self.parent_class}(",
            f"{self.parent_class}(static_cast<{parent_param_set}*>(paramSet), ",
        )

        # Add the param set to the initialiser list
        cpp_text = re.sub(
            # Driver constructor and initialiser list, all whitespace between ) and {
            r"(::" + self.device_class + r"\([^{]+\))(\s*){",
            # Insert param set after last entry in initialiser list, in between
            # match groups 1 and 2
            r"\1,\n    paramSet(paramSet)\2{",
            cpp_text,
        )

        return cpp_text

    def _convert_cpp_placeholder(self, original_cpp_text: str) -> str:
        cpp_text = original_cpp_text
        parent_param_set = get_param_set(self.parent_class)
        cpp_text = cpp_text.replace(
            f"{self.parent_class}(",
            f"{self.parent_class}(static_cast<{parent_param_set}*>(this), ",
        )
        return cpp_text

    def _convert_h(self, original_h: str) -> str:
        unwanted_strings = [
            string.strip() for string in self.define_strings + self.declaration_strings
        ]

        # Remove FIRST_*_PARAM define
        first_param_extractor = re.compile(r"")
        match = re.search(first_param_extractor, original_h)
        assert match, "Can't find first param"
        unwanted_strings += match.group()

        # Re-create text with unwanted lines ommited
        lines = original_h.splitlines()
        lines = [line for line in lines if line.strip() not in unwanted_strings]
        h_text = "\n".join(lines)

        # Add the constructor param set parameter
        h_text = h_text.replace(
            f" {self.device_class}(",
            f" {self.device_class}({self.device_class}ParamSet* paramSet, ",
        )

        # Add the include
        idx = h_text.index(f'#include "{self.parent_class}.h"')
        h_text = (
            h_text[:idx] + f'#include "{self.device_class}ParamSet.h"\n' + h_text[idx:]
        )
        # Add the param set pointer member definition
        protected_extractor = re.compile(r".*protected:")
        class_extractor = re.compile(r"class.* " + self.device_class + r" :[\s\S]* {")

        matches = re.findall(protected_extractor, h_text)
        class_str = re.findall(class_extractor, h_text)[0]
        param_set_definition = f"    {self.device_class}ParamSet* paramSet;"

        if matches:
            # Protected section exists - add param set here
            protected_str = matches[0]
            h_text = h_text.replace(
                protected_str,
                protected_str + "\n" + param_set_definition,
                1,
            )
        else:
            # Protected section does not exist - create it to add param set
            h_text = h_text.replace(
                class_str,
                class_str + "\n" + "protected:" + "\n" + param_set_definition,
            )

        # Replace FIRST_*_PARAM definition
        h_text = re.sub(
            r"(#define FIRST_\w+_PARAM) \w+",
            r"\1 paramSet->FIRST_" + self.device_class.upper() + "PARAMSET_PARAM",
            h_text,
        )

        return h_text

    def _convert_h_placeholder(self, original_h: str) -> str:
        h_text = original_h
        h_text = h_text.replace(
            f" {self.device_class} : public {self.parent_class}",
            (
                f" {self.device_class} : public {self.parent_class}ParamSet,"
                f" public {self.parent_class}"
            ),
        )
        return h_text

    @staticmethod
    def _insert_param_set_accessors(text: str, parameters: List[str]) -> str:
        parameters.append("NDArrayData")  # Special case to update NDArrayData
        for parameter in parameters:
            # Only match parameter name exactly, not others with same prefix
            text = re.sub(
                r"(\W)" + parameter + r"(\W)", r"\1paramSet->" + parameter + r"\2", text
            )
        return text


def filter_strings(strings: List[str], filters: List[str]) -> List[str]:
    return [
        string for string in strings if any(filter_ in string for filter_ in filters)
    ]


def get_param_set(driver: str):
    return "asynParamSet" if driver == "asynPortDriver" else driver + "ParamSet"



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List

from typing_extensions import Annotated

from pvi._schema_utils import as_discriminated_union, desc
from pvi.device import Device


class Access(Enum):
    """What access does the user have. One of:

    - R: Read only value that cannot be set. E.g. chipTemperature on a detector,
      or isHomed for a motor
    - W: Write only value that can be written to, but there is no current value.
      E.g. reboot on a detector, or overwriteCurrentPosition for a motor
    - RW: Read and Write value that can be written to and read back.
      E.g. acquireTime on a detector, or velocity of a motor
    """

    R = "R"  #: Read record only
    W = "W"  #: Write record only
    RW = "RW"  #: Read and write record

    def needs_read_record(self):
        return self != self.W

    def needs_write_record(self):
        return self != self.R


class DisplayForm(Enum):
    """Instructions for how a number should be formatted for display"""

    #: Use the default representation from value
    DEFAULT = "Default"
    #: Force string representation, most useful for array of bytes
    STRING = "String"
    #: Binary, precision determines number of binary digits
    BINARY = "Binary"
    #: Decimal, precision determines number of digits after decimal point
    DECIMAL = "Decimal"
    #: Hexadecimal, precision determines number of hex digits
    HEX = "Hex"
    #: Exponential, precision determines number of digits after decimal point
    EXPONENTIAL = "Exponential"
    #: Exponential where exponent is multiple of 3, precision determines number of
    #: digits after decimal point
    ENGINEERING = "Engineering"


@as_discriminated_union
@dataclass
class Producer:
    prefix: Annotated[
        str, desc("The prefix for record names created by the template file")
    ]

    def deserialize_parents(self, yaml_paths: List[Path]):
        """Deserialize yaml of parents and extract parameters"""
        raise NotImplementedError(self)

    def produce_device(self) -> Device:
        """Make signals from components"""
        raise NotImplementedError(self)

    def produce_csv(self, path: Path):
        """Make docs csv table"""
        raise NotImplementedError(self)

    def produce_records(self, path: Path):
        """Make epicsdbbuilder records"""
        raise NotImplementedError(self)

    def produce_other(self, path: Path, yaml_paths: List[Path]):
        """Make things like cpp, h files"""
        raise NotImplementedError(self)



from __future__ import annotations

import dataclasses
from ctypes import POINTER, c_char_p, c_int, c_void_p, pointer
from enum import Enum
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Tuple, Type, Union, get_type_hints

from apischema import serialize
from epicsdbbuilder import dbd, mydbstatic, records
from epicsdbbuilder.recordbase import Record
from typing_extensions import Annotated

from pvi._schema_utils import desc

# Add DBDs
dbd.InitialiseDbd()
for submodule_dbd in [
    # "asyn/asyn/devEpics/asynCalc.dbd",
    "asyn/asyn/devEpics/devAsynFloat32Array.dbd",
    "asyn/asyn/devEpics/devAsynFloat64.dbd",
    "asyn/asyn/devEpics/devAsynFloat64Array.dbd",
    "asyn/asyn/devEpics/devAsynFloat64TimeSeries.dbd",
    "asyn/asyn/devEpics/devAsynInt16Array.dbd",
    "asyn/asyn/devEpics/devAsynInt32.dbd",
    "asyn/asyn/devEpics/devAsynInt32Array.dbd",
    "asyn/asyn/devEpics/devAsynInt32TimeSeries.dbd",
    "asyn/asyn/devEpics/devAsynInt64.dbd",
    "asyn/asyn/devEpics/devAsynInt64Array.dbd",
    "asyn/asyn/devEpics/devAsynInt64Misc.dbd",
    "asyn/asyn/devEpics/devAsynInt64TimeSeries.dbd",
    "asyn/asyn/devEpics/devAsynInt8Array.dbd",
    "asyn/asyn/devEpics/devAsynOctet.dbd",
    "asyn/asyn/devEpics/devAsynOctetLs.dbd",
    "asyn/asyn/devEpics/devAsynUInt32Digital.dbd",
    "copyInfo/copyInfoApp/src/devCopyInfo.dbd",
    "busy/busyApp/src/busyRecord.dbd",
    # "busy/busyApp/src/busySupport_LOCAL.dbd",
    "busy/busyApp/src/busySupport_withASYN.dbd",
    # "busy/busyApp/src/testBusyAsynInclude.dbd",
]:
    dbd.LoadDbdFile(Path(__file__).parent.parent / "submodules" / submodule_dbd)


# https://code.activestate.com/recipes/576731-c-function-decorator/
class DbFunction:
    """This class wraps a Python function into its C equivalent in the given library"""

    def __init__(self, error_check=None):
        """Set the library to reference the function name against."""
        self.error_check = error_check

    def __call__(self, f):
        """Performs the actual function wrapping."""

        class PointerFactory:
            def __getitem__(self, key):
                """Return a pointer to the type, used in get_type_hints below"""
                # https://github.com/python/mypy/issues/7540
                return POINTER(key)

        # Get the type hints, redirecting pointer[something] to PointerFactory
        hints = get_type_hints(f, localns=dict(pointer=PointerFactory()))

        # Return the function with argtypes etc. described by the type hints
        return mydbstatic.GetDbFunction(
            f.__name__,
            argtypes=[v for x, v in hints.items() if x != "return"],
            restype=hints.get("return"),
            errcheck=self.error_check,
        )


@DbFunction()
def dbGetFieldDbfType(pdbentry: c_void_p) -> c_int:
    ...


@DbFunction(mydbstatic.auto_decode)
def dbGetFieldTypeString(dbfType: c_int) -> c_char_p:
    ...


@DbFunction(mydbstatic.auto_decode)
def dbGetDefault(pdbentry: c_void_p) -> c_char_p:
    ...


@DbFunction(mydbstatic.auto_decode)
def dbGetPrompt(pdbentry: c_void_p) -> c_char_p:
    ...


@DbFunction()
def dbGetNMenuChoices(pdbentry: c_void_p) -> c_int:
    ...


@DbFunction()
def dbGetMenuChoices(pdbentry: c_void_p) -> pointer[c_char_p]:
    ...


DBF_TYPE_MAP = dict(
    DBF_STRING=str,
    DBF_CHAR=int,
    DBF_UCHAR=int,
    DBF_SHORT=int,
    DBF_USHORT=int,
    DBF_LONG=int,
    DBF_ULONG=int,
    DBF_INT64=int,
    DBF_UINT64=int,
    DBF_FLOAT=float,
    DBF_DOUBLE=float,
    DBF_ENUM=str,
    DBF_MENU=Enum,
    DBF_DEVICE=str,
    DBF_INLINK=str,
    DBF_OUTLINK=str,
    DBF_FWDLINK=str,
    DBF_NOACCESS=None,
)

FieldList = List[Tuple[str, Type, dataclasses.Field]]


def make_fields(entry: dbd.DBEntry, record_type: str) -> FieldList:
    fields: FieldList = []
    for field_name in entry.iterate_fields():
        default = dbGetDefault(entry)
        prompt = dbGetPrompt(entry)
        field_type = DBF_TYPE_MAP[dbGetFieldTypeString(dbGetFieldDbfType(entry))]
        if field_type is Enum:
            # Make a custom enum type from the menu choices
            choices = dbGetMenuChoices(entry)
            str_choices = [choices[i].decode() for i in range(dbGetNMenuChoices(entry))]
            enum_name = record_type + field_name
            field_type = Enum(enum_name, [(c, c) for c in str_choices])  # type: ignore
        if field_type:
            fields.append(
                (
                    field_name,
                    Annotated[Union[field_type, str], desc(prompt)],
                    dataclasses.field(default=default),
                )
            )
    return fields


def make_fields_dict() -> Dict[str, FieldList]:
    fields_dict: Dict[str, FieldList] = {}
    entry = dbd.DBEntry()
    for record_type in entry.iterate_records():
        fields_dict[record_type] = make_fields(entry, record_type)
    return fields_dict


fields_dict = make_fields_dict()


class RecordPair:
    in_record_type: ClassVar[Type]
    out_record_type: ClassVar[Type]

    @classmethod
    def for_record_types(cls, in_str: str, out_str: str) -> Type[RecordPair]:
        in_record_type = dataclasses.make_dataclass(in_str, fields_dict[in_str])
        out_record_type = dataclasses.make_dataclass(out_str, fields_dict[out_str])
        namespace = dict(in_record_type=in_record_type, out_record_type=out_record_type)

        # Create combined record type for the record pair subclass
        in_fields = dict((a, (b, c)) for a, b, c, in fields_dict[in_str])
        in_fields.update(dict((a, (b, c)) for a, b, c, in fields_dict[out_str]))
        all_fields = [(a, b, c) for [a, (b, c)] in in_fields.items()]

        record_type = dataclasses.make_dataclass(
            f"{in_str}_{out_str}",
            all_fields,
        )
        subclass = type(f"{in_str}_{out_str}", (cls, record_type), namespace)

        return subclass

    def sort_records(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        record_fields: Dict[str, Any] = serialize(
            self, exclude_none=True, exclude_defaults=True
        )

        def make(cls: Type) -> Dict[str, Any]:
            fields = set(f.name for f in dataclasses.fields(cls))
            return {k: v for k, v in record_fields.items() if k in fields}

        return make(self.in_record_type), make(self.out_record_type)


# Populate records
AnalogueRecordPair = RecordPair.for_record_types("ai", "ao")
BinaryRecordPair = RecordPair.for_record_types("bi", "bo")
BusyRecordPair = RecordPair.for_record_types("bi", "busy")
LongRecordPair = RecordPair.for_record_types("longin", "longout")
MultiBitBinaryRecordPair = RecordPair.for_record_types("mbbi", "mbbo")
StringRecordPair = RecordPair.for_record_types("stringin", "stringout")
WaveformRecordPair = RecordPair.for_record_types("waveform", "waveform")

PVI_NELM = 100000


class PVIRecord(records.waveform, Record):
    def __init__(self, record: str):
        super().__init__(
            record,
            DESC="Driver PV Interface",
            DTYP="asynOctetRead",
            FTVL="CHAR",
            INP="@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))PVI_PARAM_TREE",
            NELM=PVI_NELM,
            SCAN="I/O Intr",
        )
